#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# This file is part of Rok.
#
# Copyright © 2020-2022 Arrikto Inc.  All Rights Reserved.

"""Script to authorize a ServiceAccount for an external client."""

import os
import sys
import enum
import yaml
import logging

from rok_kubernetes import config, models

from rok_tasks import frontend, question
from rok_tasks.frontend import Choice
from rok_tasks.frontend.cli import RokCLI

from rok_kubernetes.client import RoleBindingClient, AuthorizationPolicyClient

log = logging.getLogger(__name__)
fr = frontend.get_frontend()

TITLE = "Kubernetes ServiceAccount Authorizer"
DESCRIPTION = "Authorize a ServiceAccount in a Namespace"

AUTHORIZATION_POLICY = """
apiVersion: security.istio.io/v1beta1
kind: AuthorizationPolicy
metadata:
  name: %(name)s
  namespace: %(namespace)s
spec:
  action: ALLOW
  rules:
  - when:
    - key: request.headers[kubeflow-userid]
      values:
      - %(kubeflow_userid)s
"""

# Helpers
######################################################


def load_kube_config():
    # Load K8s configuration, if present
    if "KUBERNETES_SERVICE_HOST" in os.environ:
        config.load_incluster_config()
    else:
        try:
            config.load_kube_config()
        except IOError as e:
            raise RuntimeError("No K8s configuration was found: %s" % e)


def create_role_binding(rb):
    namespace = rb.metadata.namespace
    role_binding_client = RoleBindingClient()
    return role_binding_client.create(rb, namespace=namespace)


def create_authorization_policy(policy):
    policy_client = AuthorizationPolicyClient()
    namespace = policy["metadata"]["namespace"]
    return policy_client.create(policy, namespace=namespace)


class Role(enum.Enum):
    """Enumaration of supported roles."""

    VIEW = "view"
    EDIT = "edit"
    ADMIN = "admin"

    @property
    def desc(self):
        if self is self.VIEW:
            return "Viewer"
        elif self is self.EDIT:
            return "Editor"
        elif self is self.ADMIN:
            return "Administrator"
        else:
            raise RuntimeError("Invalid role option: %s" % self)


# Questions
######################################################


class ServiceAccountName(question.LineInputQuestion):
    """Question for Service Account Name."""

    name = "question/service_account_name"
    msg = "What is the name of the ServiceAccount?"
    required = True
    argument_opts = {
        "name": ["--service-account-name"],
        "metavar": "SERVICE_ACCOUNT_NAME",
        "help": "Use ServiceAccount with name %(metavar)s"
    }


class ServiceAccountNamespace(question.LineInputQuestion):
    """Question for Service Account Namespace."""

    name = "question/service_account_namespace"
    msg = "What is the namespace of the ServiceAccount?"
    required = True
    argument_opts = {
        "name": ["--service-account-namespace"],
        "metavar": "SERVICE_ACCOUNT_NAMESPACE",
        "help": "Use ServiceAccount with namespace %(metavar)s"
    }


class Role(question.MenuQuestion):
    """Question for Role."""

    name = "question/role"
    msg = "What is the role you wish to grant to the ServiceAccount?"
    required = True
    argument_opts = {
        "name": ["--role"],
        "metavar": "ROLE",
        "help": "Bind ServiceAccount with role %(metavar)s"
    }
    choices = [Choice(id=r.value, desc=r.desc, long_desc="") for r in Role]


class Namespace(question.LineInputQuestion):
    """Question for the Namespace to give access to."""

    def __init__(self, *args, **kwargs):
        super(Namespace, self).__init__(*args, **kwargs)
        sa_path = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
        if os.path.exists(sa_path):
            with open(sa_path) as f:
                self.default = f.readline()

    name = "question/namespace"
    msg = "What is the namespace you want to give access to?"
    required = True
    argument_opts = {
        "name": ["--namespace"],
        "help": "Namespace to give access to"
    }


# CLI
######################################################


class ServiceAccountAuthorizer(RokCLI):
    """Create a ServiceAccount and ServiceAccountToken."""

    IDENTIFIABLE_LABELS = {"arrikto.com/service-account-token-gen": ""}
    DEFAULT_LOGFILE = "service-account-authorizer.log"
    DEFAULT_FRONTEND = "readline"

    def initialize_parser(self):
        parser = super(ServiceAccountAuthorizer, self).initialize_parser()
        questions = [ServiceAccountName(), ServiceAccountNamespace(), Role(),
                     Namespace()]
        self.questions = questions
        question.add_questions_args(questions, parser)
        return parser

    def _ask_questions(self):
        # TODO: Ask all questions
        for q in self.questions:
            self.qctx.add_question(q)
            q.ask()
        return

    def make_role_binding(self):
        sa_name = self.qctx.questions[ServiceAccountName.name].get_answer()
        sa_namespace = (self.qctx.questions[ServiceAccountNamespace.name]
                        .get_answer())
        role = self.qctx.questions[Role.name].get_answer()
        namespace = self.qctx.questions[Namespace.name].get_answer()

        kubeflow_role = "kubeflow-%s" % role
        name = "%s-%s-%s" % (sa_namespace, sa_name, role)
        user_name = "system:serviceaccount:%s:%s" % (sa_namespace, sa_name)

        metadata = models.V1ObjectMeta(annotations={"user": user_name,
                                                    "role": role},
                                       name=name, namespace=namespace,
                                       labels=self.IDENTIFIABLE_LABELS)
        role_ref = models.V1RoleRef(api_group="rbac.authorization.k8s.io",
                                    kind="ClusterRole", name=kubeflow_role)
        subject = models.V1Subject(api_group="rbac.authorization.k8s.io",
                                   kind="User", name=user_name)
        return models.V1RoleBinding(api_version="rbac.authorization.k8s.io/v1",
                                    kind="RoleBinding", metadata=metadata,
                                    role_ref=role_ref, subjects=[subject])

    def make_authorization_policy(self):
        sa_name = self.qctx.questions[ServiceAccountName.name].get_answer()
        sa_namespace = (self.qctx.questions[ServiceAccountNamespace.name]
                        .get_answer())
        namespace = self.qctx.questions[Namespace.name].get_answer()

        name = "%s-%s" % (sa_namespace, sa_name)
        kubeflow_userid = "system:serviceaccount:%s:%s" % (sa_namespace,
                                                           sa_name)
        policy = AUTHORIZATION_POLICY % {"name": name,
                                         "namespace": namespace,
                                         "kubeflow_userid": kubeflow_userid}
        policy = yaml.safe_load(policy)
        return policy

    def main(self):
        load_kube_config()
        self._ask_questions()

        sa_name = self.qctx.questions[ServiceAccountName.name].get_answer()
        sa_namespace = (self.qctx.questions[ServiceAccountNamespace.name]
                        .get_answer())
        role = self.qctx.questions[Role.name].get_answer()
        namespace = self.qctx.questions[Namespace.name].get_answer()

        with fr.Progress(maxval=4, title=TITLE) as p:

            p.info("Generating RoleBinding...")
            rb = self.make_role_binding()
            p.inc()

            p.info("Applying RoleBinding...")
            create_role_binding(rb)
            p.inc()

            p.info("Generating AuthorizationPolicy...")
            policy = self.make_authorization_policy()
            p.inc()

            p.info("Applying AuthorizationPolicy...")
            create_authorization_policy(policy)
            p.inc()

            p.success("Successfully created necessary resources")

        msg = (u"Successfully granted `%s' access in namespace `%s' to the"
               u" ServiceAccount `%s/%s'.\n"
               % (role, namespace, sa_namespace, sa_name))

        with fr.Message(msg, title=TITLE, lvl=frontend.DEBUG,
                        ok_button=True) as m:
            m.show()


def main():
    cli = ServiceAccountAuthorizer()
    return cli.run()


if __name__ == "__main__":
    sys.exit(main())
